package dev.fitko.fitconnect.integrationtests;

import com.nimbusds.jose.jwk.JWK;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.crypto.JWKPair;
import dev.fitko.fitconnect.api.domain.model.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.event.ReplyStatus;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionState;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.event.problems.other.TechnicalError;
import dev.fitko.fitconnect.api.domain.model.reply.RepliesForPickup;
import dev.fitko.fitconnect.api.domain.model.reply.SentReply;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.sender.ReceivedReply;
import dev.fitko.fitconnect.api.domain.sender.SendableReply;
import dev.fitko.fitconnect.api.domain.sender.SendableSubmission;
import dev.fitko.fitconnect.api.domain.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectSenderException;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectSubscriberException;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.bootstrap.ClientFactory;
import dev.fitko.fitconnect.client.util.ReplyChannelKeyGenerator;
import dev.fitko.fitconnect.integrationtests.condition.EnableIfEnvironmentVariablesAreSet;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.util.Date;
import java.util.List;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.greaterThan;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.lessThan;
import static org.junit.jupiter.api.Assertions.assertThrows;

@EnableIfEnvironmentVariablesAreSet
public class ReplyWorkflowIT extends IntegrationTestBase {

    private SenderClient senderClient;
    private SubscriberClient subscriberClient;

    @BeforeEach
    public void cleanup() {
        cleanupTestSubmissions();
        ApplicationConfig config =  getConfigWithCredentialsFromEnvironment();
        senderClient = ClientFactory.getSenderClient(config);
        subscriberClient = ClientFactory.getSubscriberClient(config);
    }

    @Test
    void testBasicReplyWorkflow() throws IOException {

        // #######################
        // Online-Service
        // #######################

        final String submissionData = getResourceAsString("/submission_data.json");

        // Initial submission with new case
        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID_JSON_DATA")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .build();

        final SentSubmission sentSubmission = senderClient.send(submission);

        // reply channel encryption and decryption keys
        final JWKPair jwkPair = ReplyChannelKeyGenerator.generateKeyPair();
        final JWK privateKey = jwkPair.getPrivateKey();
        final JWK publicKey = jwkPair.getPublicKey();

        // new submission with BiDiCo
        final var newSubmissionForCase = SendableSubmission.Builder()
                .setDestination(sentSubmission.getDestinationId())
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                // Optional FIT-Connect replyChannel with existing case
                .setCase(sentSubmission.getCaseId())
                .setReplyChannel(ReplyChannel.fromFitConnect(publicKey, List.of("https://test.fitko.de")))
                .setCallback(URI.create("https://webhook.site/26f465ed-b963-40b6-bbaf-5b385933ef9a"), "secretCallbackPasswordForReplyNotifications")
                .build();

        final SentSubmission sentSubmissionForCase = senderClient.send(newSubmissionForCase);
        final SubmissionStatus sentSubmissionStatus = senderClient.getStatusForSubmission(sentSubmissionForCase);
        assertThat(sentSubmissionStatus.getStatus(), is(SubmissionState.SUBMITTED));

        // #######################
        // Government Agency
        // #######################

        // ...
        // load all submission for a case and sort DESC by submittedAt to get latest submission with most current encryption key
        // ...

        final ReceivedSubmission receivedSubmission = subscriberClient.requestSubmission(sentSubmissionForCase.getSubmissionId());
        assertThat(receivedSubmission.getSubmittedAt(), is(lessThan(new Date())));
        assertThat(receivedSubmission.getSubmittedAt(), equalTo(sentSubmissionStatus.getIssuedAt()));

        final Attachment attachment = Attachment.fromPath(Path.of("src/test/resources/attachment.txt"), "application/pdf");
        final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .addAttachment(attachment)
                .build();

        // send reply to online service
        final SentReply sentReply = subscriberClient.sendReply(sendableReply);

        final ReplyStatus replyStatus = subscriberClient.getStatusForReply(sentReply);
        assertThat(replyStatus.getStatus(), is(SubmissionState.SUBMITTED));

        // ###### Online Service  ########

        // receive reply
        final RepliesForPickup repliesForPickup = senderClient.getAvailableReplies();
        assertThat(repliesForPickup.getCount(), greaterThan(0));

        // ...
        // filter available replies
        // repliesForPickup.stream().filter(reply -> reply.getReplyId() == ?)
        // ...

        // pick up reply and accept
        final ReceivedReply receivedReply = senderClient.getReply(sentReply.getReplyId(), privateKey);
        receivedReply.acceptReply();

        assertThat(receivedReply.getDataAsString(), is(submissionData));
        assertThat(receivedReply.getAttachments().size(), is(1));
        assertThat(receivedReply.getAttachments().iterator().next().getDataAsString(), is(attachment.getDataAsString()));

    }

    @Test
    void testProcessingStandardsDoNotMatch() throws IOException {

        // Given
        final JWKPair jwkPair = ReplyChannelKeyGenerator.generateKeyPair();
        final JWK publicKey = jwkPair.getPublicKey();

        final String submissionData = getResourceAsString("/submission_data.json");

        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID_JSON_DATA")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .setReplyChannel(ReplyChannel.fromFitConnect(publicKey, List.of("https://foo.bar.baz")))
                .build();

        // Then
        assertThrows(FitConnectSenderException.class, () -> senderClient.send(submission));
    }

    @Test
    void testAcceptReply() throws IOException {

        // Given
        final JWKPair jwkPair = ReplyChannelKeyGenerator.generateKeyPair();
        final JWK privateKey = jwkPair.getPrivateKey();
        final JWK publicKey = jwkPair.getPublicKey();

        final String submissionData = getResourceAsString("/submission_data.json");
        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID_JSON_DATA")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .setReplyChannel(ReplyChannel.fromFitConnect(publicKey, List.of("https://test.fitko.de")))
                .build();

        final SentSubmission sentSubmission = senderClient.send(submission);
        final SubmissionStatus status = senderClient.getStatusForSubmission(sentSubmission);
        assertThat(status.getStatus(), is(SubmissionState.SUBMITTED));

        final ReceivedSubmission receivedSubmission = subscriberClient.requestSubmission(sentSubmission.getSubmissionId());
        final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .build();

        final SentReply sentReply = subscriberClient.sendReply(sendableReply);
        final ReplyStatus replyStatus = subscriberClient.getStatusForReply(sentReply);
        assertThat(replyStatus.getStatus(), is(SubmissionState.SUBMITTED));

        // When
        final ReceivedReply receivedReply = senderClient.getReply(sentReply.getReplyId(), privateKey);

        // Then the accepting leads to a removal of the reply
        receivedReply.acceptReply();

        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.getReply(sentReply.getReplyId(), privateKey));
        assertThat(exception.getMessage(), containsString("Reply with id " + sentReply.getReplyId() + " does not exist"));
    }

    @Test
    void testRejectReply() throws IOException {

        // Given
        final JWKPair jwkPair = ReplyChannelKeyGenerator.generateKeyPair();
        final JWK privateKey = jwkPair.getPrivateKey();
        final JWK publicKey = jwkPair.getPublicKey();

        final String submissionData = getResourceAsString("/submission_data.json");
        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID_JSON_DATA")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .setReplyChannel(ReplyChannel.fromFitConnect(publicKey, List.of("https://test.fitko.de")))
                .build();

        final SentSubmission sentSubmission = senderClient.send(submission);
        final SubmissionStatus status = senderClient.getStatusForSubmission(sentSubmission);
        assertThat(status.getStatus(), is(SubmissionState.SUBMITTED));

        final ReceivedSubmission receivedSubmission = subscriberClient.requestSubmission(sentSubmission.getSubmissionId());
        final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .build();

        final SentReply sentReply = subscriberClient.sendReply(sendableReply);
        final ReplyStatus replyStatus = subscriberClient.getStatusForReply(sentReply);
        assertThat(replyStatus.getStatus(), is(SubmissionState.SUBMITTED));

        // When
        final ReceivedReply receivedReply = senderClient.getReply(sentReply.getReplyId(), privateKey);

        // Then a rejection leads to the removal of the reply
        receivedReply.rejectReply(List.of(new TechnicalError()));

        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.getReply(sentReply.getReplyId(), privateKey));
        assertThat(exception.getMessage(), containsString("Reply with id " + sentReply.getReplyId() + " does not exist"));
    }

    @Test
    void testReplyRejectionViaSenderClient() throws IOException {

        // Given
        final JWKPair jwkPair = ReplyChannelKeyGenerator.generateKeyPair();
        final JWK publicKey = jwkPair.getPublicKey();

        final String submissionData = getResourceAsString("/submission_data.json");
        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID_JSON_DATA")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .setReplyChannel(ReplyChannel.fromFitConnect(publicKey, List.of("https://test.fitko.de")))
                .build();

        final SentSubmission sentSubmission = senderClient.send(submission);
        final SubmissionStatus status = senderClient.getStatusForSubmission(sentSubmission);
        assertThat(status.getStatus(), is(SubmissionState.SUBMITTED));

        final ReceivedSubmission receivedSubmission = subscriberClient.requestSubmission(sentSubmission.getSubmissionId());
        final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .build();

        final SentReply sentReply = subscriberClient.sendReply(sendableReply);
        final ReplyStatus replyStatus = subscriberClient.getStatusForReply(sentReply);
        assertThat(replyStatus.getStatus(), is(SubmissionState.SUBMITTED));

        // When
        senderClient.rejectReply(sentReply.getReplyId(), List.of(new TechnicalError()));

        // Then a rejection leads to the removal of the reply
        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> senderClient.getReply(sentReply.getReplyId(), jwkPair.getPrivateKey()));
        assertThat(exception.getMessage(), containsString("Reply with id " + sentReply.getReplyId() + " does not exist"));
    }


    @Test
    void testSendReplyOnSubmissionThatHasNoReplyChannel() throws IOException {

        final ApplicationConfig config =  getConfigWithCredentialsFromEnvironment();
        final SenderClient senderClient = ClientFactory.getSenderClient(config);
        final SubscriberClient subscriberClient = ClientFactory.getSubscriberClient(config);

        final String submissionData = getResourceAsString("/submission_data.json");

        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID_JSON_DATA")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .build();

        final SentSubmission sentSubmission = senderClient.send(submission);
        final SubmissionStatus status = senderClient.getStatusForSubmission(sentSubmission);
        assertThat(status.getStatus(), is(SubmissionState.SUBMITTED));

        final ReceivedSubmission receivedSubmission = subscriberClient.requestSubmission(sentSubmission.getSubmissionId());

        final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .build();

        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> subscriberClient.sendReply(sendableReply));
        assertThat(exception.getMessage(), containsString("FitConnect reply channel is mandatory for sending replies"));
    }
}

